import base64
from io import BytesIO

from PIL import Image as PILImage


def decode_base64_to_image(base64_str: str) -> PILImage.Image:
    # Convert base64 string to PIL image
    img_data = base64.b64decode(base64_str)
    return PILImage.open(BytesIO(img_data))



import uuid
from typing import Any, Dict, Iterable, TypedDict, Union, Optional

import httpx
import pydantic
from typing_extensions import Literal

from ._constants import DEFAULT_MAX_RETRIES


class BaseModel(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(extra="allow", arbitrary_types_allowed=True)


class APIException(Exception):
    def __init__(self, code: int, msg: str, context: Union[Dict, None] = None):
        self.code = code
        self.msg = msg
        self.context = context

    def __repr__(self) -> str:
        return f"{self.code}: {self.msg}"

    __str__ = __repr__


class StatusCode:
    SUCCESS = 0
    REQUIRE_AUTHENTICATION = 10001
    RESOURCE_NOT_FOUND = 131005
    BAD_REQUEST = 10400
    INVALID_ACCESS_TOKEN = 20005
    PASSTIME_ACCESS_TOKEN = 20006
    INTERNAL_ERROR = 40003
    BAD_RESPONSE = 10002
    APITimeout = 10003
    APIConnectionError = 10004


class AuthenticationRequiredException(APIException):
    def __init__(self, msg: str = "Require Authentication", context: Union[Dict, None] = None):
        super().__init__(StatusCode.REQUIRE_AUTHENTICATION, msg, context)


class HttpStatusError(APIException):
    def __init__(self, code: int, msg: str, context: Union[Dict, None] = None):
        super().__init__(code, msg, context)


class BadResponseError(APIException):
    def __init__(self, msg: str = "Bad Response", context: Union[Dict, None] = None):
        super().__init__(StatusCode.BAD_RESPONSE, msg, context)


class APITimeoutError(APIException):
    def __init__(self, msg: str = "Timeout", context: Union[Dict, None] = None):
        super().__init__(StatusCode.APITimeout, msg, context)


class APIConnectionError(APIException):
    def __init__(self, msg: str = "API Connection Error", context: Union[Dict, None] = None):
        super().__init__(StatusCode.APIConnectionError, msg, context)


class RequestOptions(TypedDict):
    headers: dict
    timeout: httpx.Timeout
    params: dict
    max_retries: int
    no_auth: bool
    raw_response: bool
    files: Dict
    data: Dict
    content: Union[bytes, str, Iterable[bytes], Iterable[str]]


class FinalRequestOptions(BaseModel):
    method: Literal["get", "post", "put", "delete", "patch"]
    url: str
    headers: dict = {}
    params: dict = {}
    max_retries: Union[int, None] = DEFAULT_MAX_RETRIES
    timeout: Union[httpx.Timeout, None] = None
    json_data: Union[Dict, None] = None
    files: Union[Dict, None] = None
    data: Union[Dict, None] = None
    content: Union[bytes, str, Iterable[bytes], Iterable[str], None] = None
    no_auth: bool = False
    raw_response: bool = False

    def get_max_retries(self, max_retries: int) -> int:
        return self.max_retries if self.max_retries is not None else max_retries



class Document(BaseModel):
    """document 由多个 chunk 组成"""

    id: str = pydantic.Field(default_factory=lambda: str(uuid.uuid4()))
    """UUID"""
    summary: str | None = None

    num_chunks: int | None = None
    child_chunk_ids: list[str] | None = None
    created_at: str | None = None
    updated_at: str | None = None
    metadata: dict | None = None

    chunks: list["Chunk"] = []

    def get_chunks(self) -> list["Chunk"]:
        flatten_chunks = []
        for chunk in self.chunks:
            if chunk.children:
                flatten_chunks.extend(chunk.children)
            else:
                flatten_chunks.append(chunk)
        return flatten_chunks

    def _add_chunk(self, chunk: "Chunk"):
        chunk.doc_id = self.id
        self.chunks.append(chunk)
        self.num_chunks = len(self.chunks)
        self.child_chunk_ids = [c.id for c in self.chunks]

    def add_chunk(self, chunk: Union["Chunk", list["Chunk"]]):
        if isinstance(chunk, list):
            for c in chunk:
                self._add_chunk(c)
        else:
            self._add_chunk(chunk)


class Chunk(BaseModel):
    """chunk 由多个 token 组成"""

    id: str = pydantic.Field(default_factory=lambda: str(uuid.uuid4()))
    """chunk UUID"""
    index: int | None = None
    """index of the chunk in the document"""
    parent_chunk_id: str | None = None
    child_chunk_ids: list[str] | None = None
    doc_id: str | None = None
    """document ID"""
    chunk_type: (
        Literal[
            "text",
            "image",
            "table_desc",
            "container",
            "markdown",
            "markdown_code",
            "markdown_title",
            "markdown_section_header",
            "table_csv",
        ]
        | str
    ) = "text"
    """chunk type
    - text: plain text
    - image: image, content is the image URL in markdown format
    - markdown: markdown content
    - table_md: markdown table
    - code: code snippet
    - table_desc: table description
    - container: container for other chunks
    """
    content: str | None = None
    """chunk content"""
    image_name: str | None = None
    image_content: str | None = None
    """encoded image content"""
    table_content: str | None = None
    """table content, decided by chunk_type"""
    num_tokens: int | None = None
    """number of tokens"""
    created_at: str | None = None
    """chunk creation time"""
    updated_at: str | None = None
    """chunk update time"""
    metadata: dict | None = None
    """Othre metadata"""

    children: list["Chunk"] = []


class AllowedExtensions(BaseModel):
    allowed_extensions: list[str]


class BaseResponse(BaseModel):
    code: int
    """错误码，非 0 表示失败"""
    msg: str
    """错误描述"""
    data: Union[Any, None] = None
    """返回数据"""
    process_time: float | None = None
    """处理时间"""

    model_config = pydantic.ConfigDict(extra="allow")


class AllowedExtensionsResponse(BaseResponse):
    data: AllowedExtensions


class DocumentResponse(BaseResponse):
    data: Document


class ParseParams(BaseModel):
    has_watermark: bool | None = None
    """是否有水印"""
    force_convert_pdf: bool | None = None
    """是否强制转换 PDF"""
    use_olm_pdf: Optional[bool] = None
    olm_api_url: Optional[str] = None
    api_key: Optional[str] = None



import httpx

# default timeout is 10 minutes
DEFAULT_TIMEOUT = httpx.Timeout(timeout=600.0, connect=5.0)
DEFAULT_MAX_RETRIES = 2
DEFAULT_CONNECTION_LIMITS = httpx.Limits(max_connections=1000, max_keepalive_connections=100)

INITIAL_RETRY_DELAY = 0.5
DEFAULT_RETRY_DELAY = 1.0
MAX_RETRY_DELAY = 8.0
DEFAULT_WRITE_ROW_BATCH_SIZE = 4000
DEFAULT_WRITE_COL_BATCH_SIZE = 90



from ._types import Chunk, Document
from .uparse import AsyncParse
from .utils import decode_base64_to_image

__all__ = [
    "AsyncParse",
    "Document",
    "Chunk",
    "decode_base64_to_image",
]



import httpx

from ._client import AsyncAPIClient
from ._constants import DEFAULT_MAX_RETRIES, DEFAULT_TIMEOUT
from ._types import AllowedExtensionsResponse, Document, DocumentResponse, ParseParams


class AsyncParse(AsyncAPIClient):
    def __init__(
        self,
        base_url="http://localhost:8000",
        timeout: httpx.Timeout = DEFAULT_TIMEOUT,
        max_retries: int = DEFAULT_MAX_RETRIES,
    ):
        super().__init__(base_url=base_url, timeout=timeout, max_retries=max_retries)
        self.parse_endpoint = "parse"

    async def allowed_extensions(self) -> list[str]:
        res = await self.get(
            self.parse_endpoint + "/allowed_extensions",
            cast_to=AllowedExtensionsResponse,
        )
        return res.data.allowed_extensions

    async def parse(
        self,
        file_path: str,
        has_watermark: bool | None = None,
        force_convert_pdf: bool | None = None,
        force_ocr_all_pages: bool | None = None,
        timeout: httpx.Timeout | None = None,
        use_olm_pdf: bool | None = None,
        olm_api_url: str | None = None,
        api_key: str | None = None,  # 新增参数，用于OLM PDF解析时的API密钥
        headers: dict | None = None,
    ) -> Document:
        params = ParseParams(
            has_watermark=has_watermark,
            force_convert_pdf=force_convert_pdf,
            force_ocr_all_pages=force_ocr_all_pages,
            use_olm_pdf=use_olm_pdf,
            olm_api_url=olm_api_url,
            api_key=api_key,
        )
        
        # 将所有非空参数放入 data 字典中
        data_dict = params.model_dump(exclude_none=True)

        options = {
            "files": {"file": open(file_path, "rb")},
            "data": data_dict,
            "timeout": timeout,
        }
        
        # 初始化headers如果未提供
        if headers is None:
            headers = {}
        
        # 当use_olm_pdf为True且提供了api_key时，添加X-API-Key头
        if use_olm_pdf and api_key:
            headers["X-API-Key"] = api_key
        
        # 如果有headers，添加到options中
        if headers:
            options["headers"] = headers
        res = await self.post(
            self.parse_endpoint,
            cast_to=DocumentResponse,
            options=options,
        )
        return res.data



import inspect
from typing import TYPE_CHECKING, Dict, Type, TypeVar, Union, cast

import anyio
import httpx
import pydantic
from httpx import URL
from loguru import logger

from ._constants import (
    DEFAULT_MAX_RETRIES,
    DEFAULT_TIMEOUT,
    INITIAL_RETRY_DELAY,
    MAX_RETRY_DELAY,
)
from ._types import (
    APIConnectionError,
    APIException,
    APITimeoutError,
    BadResponseError,
    FinalRequestOptions,
    RequestOptions,
)

if TYPE_CHECKING:
    from ._types import BaseResponse

ResponseT = TypeVar(
    "ResponseT",
    bound=Union[
        None,
        dict,
        list,
        "BaseResponse",
    ],
)


class AsyncAPIClient:
    _client: httpx.AsyncClient
    max_retries: int
    auth_headers: dict

    def __init__(
        self,
        *,
        base_url: Union[str, URL],
        max_retries: int = DEFAULT_MAX_RETRIES,
        timeout: httpx.Timeout = DEFAULT_TIMEOUT,
        proxies: Union[None, httpx._types.ProxyTypes] = None,
    ):
        self._client = httpx.AsyncClient(base_url=base_url, timeout=timeout, proxies=proxies)
        self.max_retries = max_retries

    async def get_auth_headers(self) -> dict:
        return {}

    @property
    def default_headers(self) -> dict:
        return {
            "Accept": "application/json",
        }

    async def _build_headers(self, options: FinalRequestOptions) -> httpx.Headers:
        headers = {**self.default_headers, **options.headers}
        if not options.no_auth:
            auth_headers = await self.get_auth_headers()
            headers = {**headers, **auth_headers}
        return httpx.Headers(headers)

    async def _build_request(self, options: FinalRequestOptions) -> httpx.Request:
        headers = await self._build_headers(options)
        content_type = headers.get("Content-Type")
        if options.json_data is not None and content_type is None:
            headers["Content-Type"] = "application/json; charset=utf-8"
        # if options.files is not None or options.data is not None and content_type is None:
        #     headers["Content-Type"] = "multipart/form-data"
        # httpx uses connection pooling, uncompatible with asyncio
        headers["Connection"] = "close"
        kwargs = {}
        if options.timeout is not None:
            kwargs["timeout"] = options.timeout
        return self._client.build_request(
            method=options.method,
            url=options.url,
            params=options.params,
            headers=headers,
            json=options.json_data,
            files=options.files,
            data=options.data,
            content=options.content,
            **kwargs,
        )

    def _make_status_error_from_response(self, response: httpx.Response) -> APIException:
        try:
            body = response.json()
            if "code" not in body:
                raise ValueError("No code in response")
        except Exception as e:
            return BadResponseError(msg=str(e), context={"response": response.text})
        return APIException(code=body["code"], msg=body.get("msg", ""), context=body)

    async def _retry_request(
        self,
        cast_to: Type[ResponseT],
        options: FinalRequestOptions,
        remaining_retries: int,
    ) -> ResponseT:
        remaining = remaining_retries - 1
        if remaining == 1:
            logger.debug("1 retry left")
        else:
            logger.debug(f"{remaining} retries left")
        max_retries = options.get_max_retries(self.max_retries)
        retry_timeout = min(INITIAL_RETRY_DELAY * 2 ** (max_retries - remaining), MAX_RETRY_DELAY)
        logger.info(f"Retrying {options.url} in {retry_timeout} seconds")
        await anyio.sleep(retry_timeout)

        return await self._request(
            cast_to=cast_to,
            options=options,
            remaining_retries=remaining,
        )

    def _should_retry(self, response: httpx.Response) -> bool:
        logger.debug(f"Server error {response.status_code}")
        if response.status_code >= 500:
            return True
        return False

    async def _request(
        self,
        cast_to: Type[ResponseT],
        options: FinalRequestOptions,
        remaining_retries: Union[None, int] = None,
    ) -> ResponseT:
        request = await self._build_request(options)
        retries = (
            remaining_retries
            if remaining_retries is not None
            else options.get_max_retries(self.max_retries)
        )

        try:
            response = await self._client.send(request)
        except httpx.TimeoutException as e:
            logger.debug(f"Request timed out: {e}")
            if retries > 0:
                return await self._retry_request(cast_to, options, retries)
            raise APITimeoutError(context={"request": request}) from e
        except Exception as e:
            logger.debug(f"Request failed: {e}")
            if retries > 0:
                return await self._retry_request(cast_to, options, retries)
            raise APIConnectionError(context={"request": request}) from e

        try:
            response.raise_for_status()
        except httpx.HTTPStatusError as e:
            logger.debug(f"Request failed: {e}, {response.text}")
            if retries > 0 and self._should_retry(response):
                return await self._retry_request(cast_to, options, retries)
            raise self._make_status_error_from_response(e.response) from e
        if options.raw_response:
            return cast(ResponseT, response)
        if response.json()["code"] != 200:
            raise self._make_status_error_from_response(response)
        try:
            if inspect.isclass(cast_to) and issubclass(cast_to, pydantic.BaseModel):
                return cast(ResponseT, cast_to.model_validate(response.json()))
            return cast(ResponseT, response.json())
        except Exception as e:
            logger.error(f"Encountered Bad Response: {response.json()}")
            raise BadResponseError(str(e), context={"response": response.json()})

    async def request(
        self,
        cast_to: Type[ResponseT],
        options: FinalRequestOptions,
        remaining_retries: Union[int, None] = None,
    ):
        return await self._request(cast_to, options, remaining_retries)

    async def get(
        self,
        path: str,
        *,
        cast_to: Type[ResponseT],
        options: RequestOptions = {},
    ) -> ResponseT:
        opts = FinalRequestOptions(method="get", url=path, **options)
        return await self.request(cast_to, opts)

    async def post(
        self,
        path: str,
        *,
        body: Union[Dict, None] = None,
        cast_to: Type[ResponseT],
        options: RequestOptions = {},
    ) -> ResponseT:
        opts = FinalRequestOptions(method="post", json_data=body, url=path, **options)
        return await self.request(cast_to, opts)

    async def put(
        self,
        path: str,
        *,
        body: Union[Dict, None] = None,
        cast_to: Type[ResponseT],
        options: RequestOptions = {},
    ) -> ResponseT:
        opts = FinalRequestOptions(method="put", json_data=body, url=path, **options)
        return await self.request(cast_to, opts)

    async def delete(
        self,
        path: str,
        *,
        cast_to: Type[ResponseT],
        options: RequestOptions = {},
    ) -> ResponseT:
        opts = FinalRequestOptions(method="delete", url=path, **options)
        return await self.request(cast_to, opts)

    async def patch(
        self,
        path: str,
        *,
        body: Union[Dict, None] = None,
        cast_to: Type[ResponseT],
        options: RequestOptions = {},
    ) -> ResponseT:
        opts = FinalRequestOptions(method="patch", json_data=body, url=path, **options)
        return await self.request(cast_to, opts)
